""" Basic BST code for inserting (i.e. building) and printing a tree

    Your ***second standard viva task*** (of 5) will be to implement a find method into
    the class BinaryTree from pseudocode. See the STD_2 task sheet for Week 5.

    Your ***first advanced viva task*** (of 3) will be to implement a remove (delete) method
    into the class Binary Tree from partial pseudocode. See the ADV_1 task sheet for Week 5.

    Since the given code is in python it is strongly suggested you stay with python; but
    if you want to reimplement as C++ this is also OK (see the lab sheet guidance).
"""

import math

""" Node class
"""


class Node:
    def __init__(self, data=None):
        self.data = data            #Stores the value for the node we are currently on
        self.left = None            #Stores the value of the node to the left
        self.right = None           #Stores the value of the node to the right


""" BST class with insert and display methods. display pretty prints the tree
"""


class BinaryTree:
    def __init__(self):
        self.root = None            #initialises the root node

    def insert(self, data):
        if self.root is None:
            # Assigns the root node the value inserted if the node is empty
            self.root = Node(data)
        else:
            #Runs the data given through to be inserted into the tree by another function
            self._insert(data, self.root)

    def _insert(self, data, cur_node): #Inserts the given number into the tree
        if data < cur_node.data:        #checks wheather the data given is smaller than our root/parent node
            if cur_node.left is None:
                cur_node.left = Node(data)  #Assigns the data to the node to the left of the parent if and only if it is empty
            else:
                self._insert(data, cur_node.left) #In case the left node isn't empty run the function again with the left node as the parent
        elif data > cur_node.data:      #checks wheather the data given is larger than our root/parent node
            if cur_node.right is None:
                cur_node.right = Node(data) #Assigns the data to the node to the right of the parent if and only if it is empty
            else:
                self._insert(data, cur_node.right)  #In case the right node isn't empty run the function again with the right node as the parent
        else:
            print("Value already present in tree") # Runs if the value inserted has already been added to the tree before

    ## Following is the function to search for a value in a binary tree recursively ##
    def BinTreeFindRec(self, target):
        if self.root:                                               #Checks if there is a value in the root
            if self._BinTreeFinRec(target, self.root):              #Pushes the value asked for to the sub function to be looped over the tree
                return True                                         #return True if value is found in the tree
            return False                                            #return false if value isn't found in the tree
        else:
            return None                                             #return None if there is no tree
    ## Following is the subFunction to search for a value in a binary tree recursively ##
    def _BinTreeFinRec(self, target, cur_node):
        if target > cur_node.data and cur_node.right:               #Checks if value searched for is greater than current node value #
                                                                    # and checks if there is a node to the right of the current node value
            return self._BinTreeFinRec(target, cur_node.right)      #Recursively calles itself with the target value and the node to the right of the current node value
        elif target < cur_node.data and cur_node.left:              #Checks if value searched for is less than current node value #
                                                                    # and checks if there is a node to the left of the current node value
            return self._BinTreeFinRec(target, cur_node.left)       #Recursively calles itself with the target value and the node to the left of the current node value
        if target == cur_node.data:                                 #Checks if the value searched for is equal to the node we are currently on
            return True                                             #return True if value is found

    def display(self, cur_node):            #Used to showcase the current tree
        lines, _, _, _ = self._display(cur_node) #Assigns the amount of lines to be printed as well as their dimensions and the data to be printed ie. Depth and size
        for line in lines:                  #Loops over the lines to print the dara shown
            print(line)

    def _display(self, cur_node):

        if cur_node.right is None and cur_node.left is None: # checks if given node has any children
            line = '%s' % cur_node.data                      #Assigns value to be printed as the root node on the display
            width = len(line)                                #Assigns the width to be printed based on the size of the value in the node
            height = 1                                       #Assigns the height difference between this node and leading ones after
            middle = width // 2                              #Assigns the width to be printed inbetween values
            return [line], width, height, middle             #Returns dimensions to parent function to be printed

        if cur_node.right is None:                           # Checks if the node has a child to the right
            lines, n, p, x = self._display(cur_node.left)    #
            s = '%s' % cur_node.data                         #Assigns s the value to be printed for the node
            u = len(s)
            first_line = (x + 1) * ' ' + (n - x - 1) * '_' + s #Essentially creates the structure/shape of this line that will go on to be printed
            second_line = x * ' ' + '/' + (n - x - 1 + u) * ' ' #Essentially creates the structure/shape of this line that will go on to be printed
            shifted_lines = [line + u * ' ' for line in lines]  #Essentially creates the structure/shape of this line that will go on to be printed
            return [first_line, second_line] + shifted_lines, n + u, p + 2, n + u // 2 #Returns Dimensions to psrent function to be printed

        if cur_node.left is None:                            # Checks if the node has a child to the left
            lines, n, p, x = self._display(cur_node.right)
            s = '%s' % cur_node.data                         #Assigns s the value to be printed for the node
            u = len(s)
            first_line = s + x * '_' + (n - x) * ' '    #Essentially creates the structure/shape of this line that will go on to be printed
            second_line = (u + x) * ' ' + '\\' + (n - x - 1) * ' '  #Essentially creates the structure/shape of this line that will go on to be printed
            shifted_lines = [u * ' ' + line for line in lines]  #Essentially creates the structure/shape of this line that will go on to be printed
            return [first_line, second_line] + shifted_lines, n + u, p + 2, u // 2    #Returns Dimensions to psrent function to be printed

        left, n, p, x = self._display(cur_node.left)
        right, m, q, y = self._display(cur_node.right)
        s = '%s' % cur_node.data                             #Assigns s the value to be printed for the node
        u = len(s)
        first_line = (x + 1) * ' ' + (n - x - 1) * '_' + s + y * '_' + (m - y) * ' '    #Essentially creates the structure/shape of this line that will go on to be printed
        second_line = x * ' ' + '/' + (n - x - 1 + u + y) * ' ' + '\\' + (m - y - 1) * ' '  #Essentially creates the structure/shape of this line that will go on to be printed
        # this next section determines the spaces to be printed in between nodes
        if p < q:
            left += [n * ' '] * (q - p)
        elif q < p:
            right += [m * ' '] * (p - q)
        zipped_lines = zip(left, right)
        lines = [first_line, second_line] + [a + u * ' ' + b for a, b in zipped_lines]
        return lines, n + m + u, max(p, q) + 2, n + u // 2                          #Returns Dimensions to psrent function to be printed




# example calls, which construct and display the tree
bst = BinaryTree()
bst.insert(4)
bst.insert(2)
bst.insert(6)
bst.insert(1)
bst.insert(3)
bst.insert(5)
bst.insert(7)
bst.insert(8)
bst.insert(9)
##bst.insert(10)
##bst.insert(11)
##bst.insert(12)
##bst.insert(13)
##bst.insert(14)
##bst.insert(15)
bst.insert(100)
##bst.insert(200)

bst.display(bst.root)
print(bst.BinTreeFindRec(4))
print(bst.BinTreeFindRec(2))
print(bst.BinTreeFindRec(6))
print(bst.BinTreeFindRec(1))
print(bst.BinTreeFindRec(3))
print(bst.BinTreeFindRec(5))
print(bst.BinTreeFindRec(7))
print(bst.BinTreeFindRec(8))
print(bst.BinTreeFindRec(9))
print(bst.BinTreeFindRec(12))
print(bst.BinTreeFindRec(100))

bst.display(bst.root)